import functools
import gzip
import itertools
import json
from collections import defaultdict
from pathlib import Path
from typing import Final, Iterable

import numpy as np
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from tqdm import tqdm

from . import gc_content
from .common import cosine_distance

DINUCLEOTIDE_TUPLES: Final = set((n1, n2) for n1 in "ACGT" for n2 in "ACGT")
DINUCLEOTIDES: Final = [f"{n1}{n2}" for (n1, n2) in DINUCLEOTIDE_TUPLES]


def _dinucleotide_generator(record: SeqRecord) -> Iterable[tuple[str, str]]:
    return itertools.pairwise(record.seq)


@functools.cache
def cache_dinucleotides(
    aid, path: Path, cache_dir: Path = Path("../cache/dinucleotides")
):
    cache_dir.mkdir(exist_ok=True, parents=True)
    cache_path = (cache_dir / f"{aid}-dinucleotides").with_suffix(".json")

    if cache_path.is_file():
        with cache_path.open("r") as f:
            dinucleotides_ = json.load(f)
    else:
        if path.suffix == ".gz":
            with gzip.open(path, "rt") as f:
                records = SeqIO.parse(f, "fasta")
                dinucleotides_ = _count_dinucleotides(records)
        else:
            with path.open("r") as f:
                records = SeqIO.parse(f, "fasta")
                dinucleotides_ = _count_dinucleotides(records)

        with cache_path.open("w") as f:
            json.dump(dinucleotides_, f)

    return dinucleotides_


def _count_dinucleotides(records):
    dinucleotides__ = defaultdict(int)
    for rec in tqdm(
        records,
        desc="Iterating records to calculate dinucleotides",
        position=3,
        leave=False,
    ):
        for din in tqdm(
            _dinucleotide_generator(rec),
            desc=f"Iterating dinucleotides in record {rec.id}",
            position=4,
            leave=False,
            total=len(rec.seq) - 1,
        ):
            dinucleotides__[din] += 1
    dinucleotides_ = {
        f"{din[0]}{din[1]}": dinucleotides__[din] for din in DINUCLEOTIDE_TUPLES
    }
    return dinucleotides_


def dinucleotide_odds_ratio_cosine_distance(
    left_dinucleotides, left_gc, right_dinucleotides, right_gc
):
    left_odds_ratios = np.array(
        [
            left_dinucleotides[din]
            / (
                left_gc[_nucleotide_to_index(din[0])]
                * left_gc[_nucleotide_to_index(din[1])]
            )
            for din in DINUCLEOTIDES
        ]
    )
    left_odds_ratios = np.nan_to_num(left_odds_ratios, copy=False)
    right_odds_ratios = np.array(
        [
            right_dinucleotides[din]
            / (
                right_gc[_nucleotide_to_index(din[0])]
                * right_gc[_nucleotide_to_index(din[1])]
            )
            for din in DINUCLEOTIDES
        ]
    )
    right_odds_ratios = np.nan_to_num(right_odds_ratios, copy=False)

    return cosine_distance(left_odds_ratios, right_odds_ratios)


class Dinucleotides:
    """Calculate cosine similarity between dinucleotide odds ratios

    Dinucleotide odds ratio is as defined by Karlin and Burge in https://doi.org/10.1016/S0168-9525(00)89076-9
    """

    def __init__(self, path_dict):
        """
        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    def distance(self, left: str, right: str) -> float:
        left_gc = gc_content.cache_gc_content(left, self.path_dict[left])
        left_dinucleotides = cache_dinucleotides(left, self.path_dict[left])
        right_gc = gc_content.cache_gc_content(right, self.path_dict[right])
        right_dinucleotides = cache_dinucleotides(right, self.path_dict[right])

        return dinucleotide_odds_ratio_cosine_distance(
            left_dinucleotides, left_gc, right_dinucleotides, right_gc
        )

    def transform(self, aid: str) -> np.ndarray:
        """Transforms the aid into the corresponding array.

        Parameters
        ----------
        aid: str
            Aid to transform

        Returns
        -------
        np.ndarray
            Array of dinucleotide odds ratios, in a deterministic order
        """
        gc = gc_content.cache_gc_content(aid, self.path_dict[aid])
        dinucleotides_ = cache_dinucleotides(aid, self.path_dict[aid])
        transform_ = np.array(
            [
                dinucleotides_[din]
                / (gc[_nucleotide_to_index(din[0])] * gc[_nucleotide_to_index(din[1])])
                for din in DINUCLEOTIDES
            ]
        )
        transform_ = np.nan_to_num(transform_, copy=False)
        return transform_


def _nucleotide_to_index(nucleotide: str) -> int:
    if nucleotide == "A":
        return 0
    elif nucleotide == "C":
        return 1
    elif nucleotide == "G":
        return 2
    else:
        return 3



import functools
import logging
import os
import urllib.request
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Final, Optional

import pandas as pd
from ete3 import NCBITaxa

Metadata = dict[str, dict[str, Any]]
Taxonomy = list[dict[str, str]]

SOUGHT_RANKS: Final = [
    "superkingdom",
    "realm",
    "kingdom",
    "phylum",
    "subphylum",
    "class",
    "order",
    "suborder",
    "family",
    "subfamily",
    "genus",
    "subgenus",
    "species",
    "subspecies",
    "strain",
]

ICTV_SOUGHT_RANK_NAMES: Final = [
    "Superkingdom",
    "Realm",
    "Kingdom",
    "Phylum",
    "Subphylum",
    "Class",
    "Order",
    "Suborder",
    "Family",
    "Subfamily",
    "Genus",
    "Subgenus",
    "Species",
    "Subspecies",
    "Strain",
    "Genome composition",
    "Host/Source",
]


@functools.lru_cache()
def parse_ictv(ictv_vmr_path: Path = None) -> Optional[pd.DataFrame]:
    if ictv_vmr_path is not None:
        return pd.read_excel(ictv_vmr_path)
    else:
        _ictv_vmr_dir = Path(os.path.realpath(__file__)).parent

        _ictv_vmr_path = next(
            (
                path
                for path in _ictv_vmr_dir.iterdir()
                if path.suffix == ".xlsx" and "VMR" in path.stem
            ),
            None,
        )
        if _ictv_vmr_path is not None:
            try:
                return pd.read_excel(_ictv_vmr_path)
            except FileNotFoundError as e:

                logging.debug("ICTV VMR not found in expected place.")

                _ictv_vmr_df = None
        else:
            url = "https://ictv.global/filebrowser/download/461"

            now = datetime.now().strftime("%Y-%m-%d")
            urllib.request.urlretrieve(url, _ictv_vmr_dir / f"VMR-{now}.xlsx")
            return parse_ictv()


def get_ictv_taxonomy(
    species: str, aid: str, ictv_vmr_path: Path = None
) -> Optional[dict[str, str]]:
    """Gets the taxonomy for the virus species from ictv.

    Parameters
    ----------
    species: str
        Species name
    aid: str
        Aid of sequence.
    ictv_vmr_path: Path
        Path to ictv vmr file.

    Returns
    -------
    taxonomy: Optional[Dict[str, str]]
        The taxonomic ranks for the species, as {"species": species, "family": family"} etc.

    """
    if species is None:
        return None

    ictv_vmr_df = parse_ictv(ictv_vmr_path)
    if ictv_vmr_df is None:
        raise FileNotFoundError("ICTV VMR not found")

    aid_ = aid.split(".")[0]
    df = ictv_vmr_df[
        (ictv_vmr_df["Species"] == species)
        | (ictv_vmr_df["Virus name(s)"] == species)
        | (ictv_vmr_df["Virus GENBANK accession"].str.contains(aid_))
        | (ictv_vmr_df["Virus REFSEQ accession"].str.contains(aid_))
    ]
    if len(df) == 0:
        logging.warning(f"Aid {aid} ({species}) not found in ICTV taxonomy")
        return None
    else:
        return defaultdict(
            str,
            {
                rank.lower(): df[rank].values[0]
                for rank in ICTV_SOUGHT_RANK_NAMES
                if rank in df.columns
            },
        )

@functools.lru_cache(128)
def get_ncbi_taxonomy(species: str) -> Optional[dict[str, str]]:
    """Gets the taxonomy for the species from ncbi.

    Parameters
    ----------
    species: str
        Species name

    Returns
    -------
    taxonomy: Optional[Dict[str, str]]
        The taxonomic ranks for the species, as {"species": species, "family": family"} etc.

    """
    if species is None:
        return None

    ncbi = NCBITaxa()
    name_translator = ncbi.get_name_translator([species])
    tax = _get_single_name_taxonomy(species, name_translator, ncbi)
    if tax is None:
        return None
    else:
        return parse_taxonomy(tax)


def _get_single_name_taxonomy(
    species: str, name_translator: dict[str, str], ncbi: NCBITaxa
) -> Optional[Taxonomy]:
    if species not in name_translator:
        return None
    taxid = name_translator[species][0]
    lineage = ncbi.get_lineage(taxid)
    ranks = ncbi.get_rank(lineage)
    names = ncbi.get_taxid_translator(lineage)
    return [{"Rank": ranks[k], "ScientificName": names[k]} for k in lineage]


def parse_taxonomy(taxonomy: Taxonomy) -> dict[str, str]:
    """Parse out relevant taxonomic classes from the raw taxonomy.

    Parameters
    ----------
    taxonomy : List[Dict[str, str]]
        Raw taxonomy, as retrieved from the Taxonomy db

    Returns
    -------
    Dict[str, str]
        Dictionary with the ranks (e.g. "superkingdom") as keys
    """
    if taxonomy is None:
        return defaultdict(str)

    parsed_tax = {
        tax["Rank"]: tax["ScientificName"]
        for tax in taxonomy
        if tax["Rank"] in SOUGHT_RANKS
    }

    return defaultdict(str, {rank: parsed_tax.get(rank, "") for rank in SOUGHT_RANKS})



import functools
import gzip
import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, Final, Iterable, NewType

import numba
import numpy as np

# noinspection PyPackageRequirements
from Bio import SeqIO

# noinspection PyPackageRequirements
from Bio.SeqRecord import SeqRecord

from .common import cosine_distance

CodonTable = NewType("CodonTable", dict[str, float])
RSCUTable = NewType("RSCUTable", dict[str, float])
RelativeAdaptivenessTable = NewType("RelativeAdaptivenessTale", dict)


def generate_relative_adaptiveness_table(
    rscu_table: RSCUTable,
) -> RelativeAdaptivenessTable:
    """Generates the relative adaptiveness table as specified by Sharp and Li, termed $w$ by them.

    For each codon this is defined as the codon's relative synonymous codon usage divided by the maximum
    relative synonymous codon usage of the synonymous codons.  $w_{ij} = RSCU_{ij} / max_l RSCU{il}$

    Parameters
    ----------
    rscu_table: RSCUTable
        The relative synonymous codon usage table.

    Returns
    -------
    RelativeAdaptivenessTable
        The relative adaptiveness table.

    References
    ----------
    [1] https://doi.org/10.1093/nar/15.3.1281
    """
    rat = defaultdict(float)

    for synonymous_codons in _SYNONYMOUS_CODONS.values():
        rcsu_max = max(rscu_table[codon] for codon in synonymous_codons)
        for codon in synonymous_codons:
            rat[codon] = rscu_table[codon] / rcsu_max

    return rat


def generate_rscu_table(codon_count: CodonTable) -> RSCUTable:
    """Generates the Relative synonymous codon usage from the codon counts.

    Defined for each codon as the codon count over the average codon count.

    Parameters
    ----------
    codon_count: CodonTable
        The counts of each codon.

    Returns
    -------
    RSCUTable
        The relative synonymous codon usage of each codon.

    References
    ----------
    [1] https://doi.org/10.1093/nar/15.3.1281
    """
    rcsu = defaultdict(float)

    for synonymous_codons in _SYNONYMOUS_CODONS.values():
        expected_frequency = sum(
            codon_count[codon] for codon in synonymous_codons
        ) / len(synonymous_codons)

        for codon in synonymous_codons:
            rcsu[codon] = codon_count[codon] / expected_frequency

    return rcsu


@functools.cache
def generate_rscu_array(aid: str, path: Path) -> np.ndarray:
    codon_count = cache_codon_count(aid, path)
    rscu = generate_rscu_table(codon_count)

    return np.array([rscu[codon] for codon in VALID_CODONS])


def count_codons(records: Iterable[SeqRecord]) -> CodonTable:
    """Counts the codon in the records.  Each sequence is assumed to be in-frame.

    Parameters
    ----------
    records: Iterable[SeqRecord]
        All records that the codon usage is calculated for.

    Returns
    -------
    CodonTable
        Count of each codon.

    References
    ----------
    [1] https://doi.org/10.1093/nar/15.3.1281
    """
    codon_count = defaultdict(float)
    for record in records:
        for codon in _codon_generator(record):
            codon_count[codon] += 1

    # From Sharp and Li:
    # "Note that if a certain codon is never used in the reference set
    # then the CAI for any other gene in which that codon appears becomes zero.
    # To overcome this problem we assign a value of 0.5 to any X
    # that would otherwise be zero."
    for codon in VALID_CODONS:
        if codon_count[codon] == 0:
            codon_count[codon] = 0.5

    return codon_count


_illegal_codons_discovered = defaultdict(set)


def _codon_generator(record: SeqRecord) -> Iterable[str]:
    sequence = str(record.seq)
    for codon_start in range(0, len(sequence), 3):
        codon_end = codon_start + 3
        codon = sequence[codon_start:codon_end]
        if codon not in VALID_CODONS:
            if codon not in _illegal_codons_discovered[record.id]:
                logging.warning(f"Illegal codon {codon} in gene: {record.id}")
                _illegal_codons_discovered[record.id].add(codon)

            continue

        yield codon


def codon_adaptation_index_from_counts(rat, codon_counts):
    """Calculates the codon adaptation index for a codon counts.

    Defined as the geometric mean of the relative adaptiveness of the codons.

    Parameters
    ----------
    rat: RelativeAdaptivenessTable
        The relative adaptiveness table to use as reference.
    codon_counts: Dict[str, int]
        The codons to calculate codon adaptation for.

    Returns
    -------
    float
        The codon adaptation index.

    References
    ----------
    [1] https://doi.org/10.1093/nar/15.3.1281
    """

    n_codons = 0
    cai_log_sum = 0.0

    for codon, count in codon_counts.items():
        # Exclude these codons, as stated by Sharp and Li:
        # "... the number of AUG and UGG codons are subtracted
        # from L, since the RSCU value, for AUG and UGG are both
        # fixed at 1.0, and so do not contribute to the CAI."
        if codon in {"ATG", "TGG"}:
            continue

        if rat[codon] != 0:
            cai_log_sum += np.log(rat[codon]) * count
            n_codons += count

    return np.exp(cai_log_sum / n_codons)


@numba.njit
def _codon_adaptation_index_from_counts(
    rat_array: np.ndarray, codon_counts_array: np.ndarray
):
    n_codons = 0
    cai_log_sum = 0.0
    for codon_i, count in enumerate(codon_counts_array):
        if rat_array[codon_i] != 0:
            cai_log_sum += np.log(rat_array[codon_i]) * count
            n_codons += count

    return np.exp(cai_log_sum / n_codons)


def codon_adaptation_index(
    rat: RelativeAdaptivenessTable, records: Iterable[SeqRecord]
) -> float:
    """Calculates the codon adaptation index for a given set of records, assumed to be in-frame.

    Defined as the geometric mean of the relative adaptiveness of the codons used in the records .

    Parameters
    ----------
    rat: RelativeAdaptivenessTable
        The relative adaptiveness table to use as reference.
    records: Iterable[SeqRecord]
        The records to calculate CAI for.

    Returns
    -------
    float
        The codon adaptation index for the records.

    References
    ----------
    [1] https://doi.org/10.1093/nar/15.3.1281
    """

    n_codons = 0
    cai_log_sum = 0.0

    for record in records:
        for codon in _codon_generator(record):
            # Exclude these codons, as stated by Sharp and Li:
            # "... the number of AUG and UGG codons are subtracted
            # from L, since the RSCU value, for AUG and UGG are both
            # fixed at 1.0, and so do not contribute to the CAI."
            if codon in {"ATG", "TGG"}:
                continue

            if rat[codon] != 0:
                cai_log_sum += np.log(rat[codon])
                n_codons += 1

    return np.exp(cai_log_sum / n_codons)


@functools.cache
def cache_codon_count(
    aid: str, record_path: Path, cache_dir: Path = Path("../cache/codon")
) -> CodonTable:
    """Caches the codon counts, or fetches them from disk if previously calculated.

    Parameters
    ----------
    aid: str
        Unique identifier for the records.
    record_path: Path
        Path to record.
    cache_dir: Path
        Path to where codon counts are cached.

    Returns
    -------
    CodonTable
        Count of codons in record_path.
    """
    cache_dir.mkdir(exist_ok=True, parents=True)
    cache_path = (cache_dir / f"{aid}-codons").with_suffix(".json")

    if cache_path.is_file():
        with cache_path.open("r") as f:
            codon_count = json.load(f)
    else:
        logging.info(f"Counting codons for {aid}")
        if record_path.suffix == ".gz":
            with gzip.open(record_path, "rt") as f:
                records = SeqIO.parse(f, "fasta")
                codon_count = count_codons(records)
        else:
            with record_path.open("r") as f:
                records = SeqIO.parse(f, "fasta")
                codon_count = count_codons(records)

        with cache_path.open("w") as f:
            json.dump(codon_count, f)

    return codon_count


@functools.cache
def cache_codon_count_array(aid: str, path: Path) -> np.ndarray:
    codon_counts = cache_codon_count(aid, path)
    return np.array(
        [codon_counts[codon] for codon in VALID_CODONS if codon not in EXCLUDE_CODONS]
    )


@functools.cache
def cache_rat_array(aid: str, path: Path) -> np.ndarray:
    codon_count = cache_codon_count(aid, path)
    rscu_table = generate_rscu_table(codon_count)
    rat = generate_relative_adaptiveness_table(rscu_table)

    return np.array(
        [rat[codon] for codon in VALID_CODONS if codon not in EXCLUDE_CODONS]
    )


class CAI:
    """
    Class for calculating the codon adaptation distance. Assumes all sequences given are in-frame.

    As defined by Sharp and Li in https://doi.org/10.1093/nar/15.3.1281.

    Most of this exists in the biopython package, but since it raises an error
    on incomplete codons, we can't use it.  This would be reasonable, but I'm putting
    that responsibility on the user instead.  Specifically, it's a problem since the CDS
    from genbank sometimes is incomplete and contains a lot of N characters.
    Rather than going through and checking every gene, we're assuming the general codon usage is the
    still preserved.

    Distance here is 1 - CAI.  So a distance of 0 means a perfect CAI and adaption to the host.

    """

    def __init__(self, path_dict):
        """
        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    def distance(self, left: str, right: str) -> float:
        rat_array = cache_rat_array(right, self.path_dict[right])

        codon_count = cache_codon_count_array(left, self.path_dict[left])

        return 1 - _codon_adaptation_index_from_counts(rat_array, codon_count)

    def transform(self, aid):
        return aid


class RSCUCorrelation:
    """
    Class for distances based on correlation of Relative Synonymous Codon Usage. Assumes all sequences given are in-frame.

    Relative synonymous codon usage is as defined by Sharp and Li in https://doi.org/10.1093/nar/15.3.1281.

    Distance is (1 - RSCU correlation) / 2

    """

    def __init__(self, path_dict: Dict[str, Path] = None):
        """
        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    @functools.singledispatchmethod
    def distance(self, left, right):
        return 0.0

    @distance.register
    def _(self, left: str, right: str):
        if right not in self.path_dict or self.path_dict[right] is None:
            logging.info(f"{right} not in path_dict, or is None")
            return np.inf
        elif left not in self.path_dict or self.path_dict[left] is None:
            logging.info(f"{left} not in path_dict, or is None")
            return np.inf

        right_rscu_array = generate_rscu_array(right, self.path_dict[right])
        left_rscu_array = generate_rscu_array(left, self.path_dict[left])

        return self.distance(left_rscu_array, right_rscu_array)

    @distance.register
    def _(self, left: Path, right: Path):
        right_rscu_array = generate_rscu_array(right.stem, right)
        left_rscu_array = generate_rscu_array(left.stem, left)

        return self.distance(left_rscu_array, right_rscu_array)

    @distance.register(dict)
    def _(self, left_rscu_table: RSCUTable, right_rscu_table: RSCUTable):

        left_rscu_array = np.array([left_rscu_table[codon_] for codon_ in VALID_CODONS])
        right_rscu_array = np.array(
            [right_rscu_table[codon_] for codon_ in VALID_CODONS]
        )

        return cosine_distance(left_rscu_array, right_rscu_array)

    @distance.register
    def _(self, left_rscu_array: np.ndarray, right_rscu_array: np.ndarray):
        return cosine_distance(left_rscu_array, right_rscu_array)

    def transform(self, aid: str) -> np.ndarray:
        """Transforms the aid into the corresponding array.

        Parameters
        ----------
        aid: str
            Aid to transform

        Returns
        -------
        np.ndarray
            Array of relative synonymous codon usage, in a deterministic order
        """
        return generate_rscu_array(aid, self.path_dict[aid])


_SYNONYMOUS_CODONS: Final = {
    "CYS": ["TGT", "TGC"],
    "ASP": ["GAT", "GAC"],
    "SER": ["TCT", "TCG", "TCA", "TCC", "AGC", "AGT"],
    "GLN": ["CAA", "CAG"],
    "MET": ["ATG"],
    "ASN": ["AAC", "AAT"],
    "PRO": ["CCT", "CCG", "CCA", "CCC"],
    "LYS": ["AAG", "AAA"],
    "STOP": ["TAG", "TGA", "TAA"],
    "THR": ["ACC", "ACA", "ACG", "ACT"],
    "PHE": ["TTT", "TTC"],
    "ALA": ["GCA", "GCC", "GCG", "GCT"],
    "GLY": ["GGT", "GGG", "GGA", "GGC"],
    "ILE": ["ATC", "ATA", "ATT"],
    "LEU": ["TTA", "TTG", "CTC", "CTT", "CTG", "CTA"],
    "HIS": ["CAT", "CAC"],
    "ARG": ["CGA", "CGC", "CGG", "CGT", "AGG", "AGA"],
    "TRP": ["TGG"],
    "VAL": ["GTA", "GTC", "GTG", "GTT"],
    "GLU": ["GAG", "GAA"],
    "TYR": ["TAT", "TAC"],
}

VALID_CODONS: Final = {
    "TGT",
    "TGC",
    "GAT",
    "GAC",
    "TCT",
    "TCG",
    "TCA",
    "TCC",
    "AGC",
    "AGT",
    "CAA",
    "CAG",
    "ATG",
    "AAC",
    "AAT",
    "CCT",
    "CCG",
    "CCA",
    "CCC",
    "AAG",
    "AAA",
    "TAG",
    "TGA",
    "TAA",
    "ACC",
    "ACA",
    "ACG",
    "ACT",
    "TTT",
    "TTC",
    "GCA",
    "GCC",
    "GCG",
    "GCT",
    "GGT",
    "GGG",
    "GGA",
    "GGC",
    "ATC",
    "ATA",
    "ATT",
    "TTA",
    "TTG",
    "CTC",
    "CTT",
    "CTG",
    "CTA",
    "CAT",
    "CAC",
    "CGA",
    "CGC",
    "CGG",
    "CGT",
    "AGG",
    "AGA",
    "TGG",
    "GTA",
    "GTC",
    "GTG",
    "GTT",
    "GAG",
    "GAA",
    "TAT",
    "TAC",
}

EXCLUDE_CODONS: Final = {"ATG", "TGG"}



__version__ = "0.1.0"

from . import (
    sequences,
    common,
    relations,
    taxonomy,
    virus_hosts,
    gc_content,
    codon,
    codon_pair,
    dinucleotides,
)
from Bio import Entrez
import os


def set_entrez_api_key(email: str = None, api_key: str = None):
    """Sets Entrez email and api key.

    The values are read from the ENV variables "ENTREZ_EMAIL" and "ENTREZ_API_KEY" if not explicitly given.

    For more information about the API, see
    https://www.ncbi.nlm.nih.gov/books/NBK25497/#chapter2.Usage_Guidelines_and_Requiremen

    Parameters
    ----------
    email: str
        Email address of the NCBI account
    api_key : str
        API key of the NCBI account

    """
    if email is None:
        Entrez.email = os.environ["ENTREZ_EMAIL"]
    else:
        Entrez.email = email

    if api_key is None:
        Entrez.api_key = os.environ["ENTREZ_API_KEY"]
    else:
        Entrez.api_key = api_key



import functools
import uuid
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set

from . import common
from .common import RANKS


@dataclass(frozen=True)
class Relations:
    signatures: tuple
    metadata: Dict[str, Dict[str, Any]]
    virus_hosts: Dict[str, List[Dict[str, str]]]
    hosts: bool
    unique_id: int = field(default_factory=uuid.uuid4)

    def __hash__(self):
        # This isn't important, needed only to allow for caching of the results
        return hash(self.unique_id)

    @functools.lru_cache()
    def index_of_clade_relations(self, virus: str, rank: str) -> Set[int]:
        virus_species = self.metadata[virus]["species"]

        if self.hosts:
            return set(
                i
                for (i, sig) in enumerate(self.signatures)
                if any(
                    self.metadata[sig].get(rank, sig) == hosts.get(rank, "NONE")
                    for hosts in self.virus_hosts[virus_species]
                )
            )
        else:
            return set(
                i
                for (i, sig) in enumerate(self.signatures)
                if self.metadata[virus].get(rank, virus)
                == self.metadata[sig].get(rank, "NONE")
            )

    @functools.lru_cache()
    def virus_relation(self, from_virus: str, to_virus: str) -> Optional[str]:
        for rank in RANKS:
            if (
                self.metadata[from_virus][rank] == self.metadata[to_virus][rank]
                and self.metadata[from_virus][rank] != ""
            ):
                return rank

        return "None"

    @functools.lru_cache()
    def host_relation(self, from_virus: str, to_host: str) -> Optional[str]:
        virus_species = self.metadata[from_virus]["species"]

        for rank in RANKS:
            for host_tax in self.virus_hosts[virus_species]:
                if self.metadata[to_host].get(rank, None) == host_tax[rank]:
                    return rank

        return "None"

    @functools.lru_cache()
    def relation(self, from_virus: str, to: str) -> Optional[str]:
        if self.metadata.get(to, None) is None:
            raise ValueError(f"Given {to} doesn't exist in metadata.")
        elif self.metadata[to]["superkingdom"] == "viruses":
            return self.virus_relation(from_virus, to)
        else:
            return self.host_relation(from_virus, to)


@functools.lru_cache(1)
def get_virus_relations():
    viruses = common.get_viruses()
    virus_metadata = common.get_virus_metadata(viruses)
    virus_hosts = common.get_virus_hosts(virus_metadata)

    return Relations(tuple(viruses), virus_metadata, virus_hosts, False)


@functools.lru_cache(1)
def with_metadata(new_metadata: Dict[str, Dict[str, Any]]):
    viruses = common.get_viruses()
    virus_metadata = common.get_virus_metadata(viruses)
    virus_hosts = common.get_virus_hosts(virus_metadata)

    return Relations(
        tuple(viruses), {**virus_metadata, **new_metadata}, virus_hosts, False
    )


def get_index_of_clade_relations(virus: str, rank: str) -> Set[int]:
    relations = get_virus_relations()
    return relations.index_of_clade_relations(virus, rank)


def virus_relation(from_virus: str, to_virus: str) -> Optional[str]:
    relations = get_virus_relations()
    return relations.virus_relation(from_virus, to_virus)


def host_relation(from_virus: str, to_host: str) -> Optional[str]:
    relations = get_virus_relations()
    return relations.host_relation(from_virus, to_host)


def relation(from_virus: str, to: str) -> Optional[str]:
    relations = get_virus_relations()
    return relations.relation(from_virus, to)



import functools
import gzip
import json
from collections import Counter
from pathlib import Path

import numba
import numpy as np
from Bio import SeqIO

from .common import cosine_distance


@functools.cache
def cache_gc_content(
    aid, path: Path, cache_dir: Path = Path("../cache/gc-content")
) -> np.ndarray:
    cache_dir.mkdir(exist_ok=True, parents=True)
    cache_path = (cache_dir / f"{aid}-gc").with_suffix(".json")

    if cache_path.is_file():
        with cache_path.open("r") as f:
            gc = json.load(f)
    else:
        if path.suffix == ".gz":
            with gzip.open(path, "rt") as f:
                records = SeqIO.parse(f, "fasta")
                seq = "".join([str(rec.seq) for rec in records])
        else:
            with path.open("r") as f:
                records = SeqIO.parse(f, "fasta")
                seq = "".join([str(rec.seq) for rec in records])

        gc = Counter(seq)
        with cache_path.open("w") as f:
            json.dump(gc, f)

    gc_ = np.array([gc.get(c, gc.get(c.lower(), 0)) for c in "ACGT"])
    return gc_


@numba.njit
def get_gc_diff(left_gc, right_gc):
    gc_i_a = left_gc / np.sum(left_gc)

    gc_j_a = right_gc / np.sum(right_gc)

    return cosine_distance(gc_i_a, gc_j_a)


class GC:
    """
    Calculate GC distance angular cosine distance
    """

    def __init__(self, path_dict):
        """
        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    def distance(self, left: str, right: str) -> float:
        left_gc = cache_gc_content(left, self.path_dict[left])
        right_gc = cache_gc_content(right, self.path_dict[right])

        return get_gc_diff(left_gc, right_gc)

    def transform(self, aid: str) -> np.ndarray:
        """Transforms the aid into the corresponding array.

        Parameters
        ----------
        aid: str
            Aid to transform

        Returns
        -------
        np.ndarray
            Array of gc content, in a deterministic order
        """
        gc_array = cache_gc_content(aid, self.path_dict[aid])
        gc_array = gc_array / np.sum(gc_array)
        return gc_array



import functools
import gzip
import itertools
import json
import logging
import os
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Final, Iterable, List, Optional, Tuple

import numba
import numpy as np
from Bio import SeqIO, SeqUtils
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from . import sequences, taxonomy, virus_hosts
from .taxonomy import Metadata

RANKS: Final[List[str]] = [
    "species",
    "genus",
    "family",
    "order",
    "class",
    "phylum",
    "kingdom",
    "superkingdom",
]

ranks_and_none = RANKS + ["None"]


if "DOWNLOAD_ROOT" in os.environ:
    DOWNLOAD_ROOT: Final[Path] = Path(os.environ["DOWNLOAD_ROOT"])
else:
    DOWNLOAD_ROOT: Final[Path] = Path.home() / "data" / "ftp.ncbi.nlm.nih.gov"

if "DOWNLOAD_FOLDER" in os.environ:
    DOWNLOAD_FOLDER: Final[Path] = Path(os.environ["DOWNLOAD_FOLDER"])
else:
    DOWNLOAD_FOLDER: Final[Path] = Path.home() / "data" / "downloads"

if "VIRUSLIST" in os.environ:
    virus_list_path: Final[Path] = Path(os.environ["VIRUSLIST"])
else:
    virus_list_path: Final[Path] = None


def chunks(iterable, size=10):
    iterator = iter(iterable)
    for first in iterator:
        yield itertools.chain([first], itertools.islice(iterator, size - 1))


@functools.lru_cache()
def get_directory(name: str):
    now = datetime.now().strftime("%Y-%m-%d")
    out_dir = Path("..") / "results" / name / now
    out_dir.mkdir(exist_ok=True, parents=True)
    return out_dir


def get_hosts() -> List[Dict[str, str]]:
    assemblies_path: Final[Path] = (
        Path(os.path.realpath(__file__)).parent.parent / "assembly_ids.json"
    )

    with assemblies_path.open("r") as f:
        hosts: List[Dict[str, str]] = json.load(f)

    hosts = [organism for organism in hosts if organism["aid"] != ""]
    unique_hosts = list({organism["aid"]: organism for organism in hosts}.values())

    return hosts


def to_fs_path(ftp_path: str, sequence_part: str = "coding") -> Optional[Path]:
    if sequence_part not in ["coding", "full-genome"]:
        raise ValueError("sequence_part has invalid value")

    path = DOWNLOAD_ROOT / Path(ftp_path)
    if not path.is_dir():
        logging.debug(f"{path} not found")
        return None

    version_path = next(path.iterdir())
    file_path = next(
        (path for path in version_path.iterdir() if _valid_path(path, sequence_part)),
        None,
    )
    return file_path


def _valid_path(path: Path, sequence_part: str) -> bool:
    if sequence_part == "coding":
        return "cds" in path.name
    else:
        return True


def parse_to_ftp_paths(hosts: List[Dict[str, Any]]) -> List[str]:
    return [to_ftp_path(organism["aid"]) for organism in hosts]


def to_ftp_path(aid: str, root: str = "genomes/all") -> str:
    db, number_and_version = aid.split("_")
    number, version = number_and_version.split(".")
    return f"{root}/{db}/{number[0:3]}/{number[3:6]}/{number[6:9]}"


def get_viruses() -> List[str]:
    virus_path = virus_list_path
    if virus_path is None or not virus_path.is_file():
        virus_path_ = Path.cwd().parent / "viruslist.txt"
        if virus_path_.is_file():
            virus_path = virus_path_

    viruses = virus_path.open("r").read().splitlines()
    return viruses


mononegavirales = [
    "Artoviridae",
    "Bornaviridae",
    "Filoviridae",
    "Lispiviridae",
    "Mymonaviridae",
    "Nyamiviridae",
    "Paramyxoviridae",
    "Pneumoviridae",
    "Rhabdoviridae",
    "Sunviridae",
    "Xinmoviridae",
]


def get_virus_hosts(virus_metadata: Metadata) -> Dict[str, List[Dict[str, str]]]:
    logging.info("Getting Hosts...")

    hosts_path = (
        Path(os.path.realpath(__file__)).parent.parent.parent
        / "cache"
        / "virus_hosts.json"
    )

    if hosts_path.exists():
        return json.load(hosts_path.open("r"))

    tax_ids = [
        (-1, t.get("species", None), t.get("aid", None))
        for t in virus_metadata.values()
    ]

    hosts = virus_hosts.get_virus_hosts(tax_ids)
    hosts = {
        key: [
            val
            for val in [taxonomy.get_ncbi_taxonomy(h) for h in hs]
            if val is not None
        ]
        for key, hs in hosts.items()
    }

    json.dump(hosts, hosts_path.open("w"))

    return hosts


def virus_row(aid: str, virus_metadata: Dict[str, Dict[str, str]]) -> Dict[str, Any]:
    return {
        "aid": aid,
        "virus": virus_metadata[aid]["species"],
        "species": virus_metadata[aid]["species"],
        "virus_genus": virus_metadata[aid]["genus"],
        "genus": virus_metadata[aid]["genus"],
        "virus_family": virus_metadata[aid]["family"],
        "family": virus_metadata[aid]["family"],
        "virus_order": virus_metadata[aid]["order"],
        "baltimore": virus_metadata[aid]["genome composition"],
        "sequence_length": _sequence_length(aid),
        "log_sequence_length": np.log10(_sequence_length(aid)),
        "average_gc_content": _gc_content(aid),
        "host_group": virus_metadata[aid]["host/source"],
    }


def _sequence_length(aid: str) -> int:
    record = sequences.get_sequence(
        aid, DOWNLOAD_FOLDER, sequences.SequencePart.DustMasked
    )
    if record is None:
        return 0
    else:
        return len(record.seq)


def _gc_content(aid: str) -> int:
    record = sequences.get_sequence(
        aid, DOWNLOAD_FOLDER, sequences.SequencePart.DustMasked
    )
    if record is None:
        return 0
    else:
        return SeqUtils.GC(record.seq) / 100


def clean_name(name: str) -> str:
    return (
        str(name).replace("\\", "").replace("(", "").replace(")", "").replace("/", "")
    )


def parse_genome_composition(composition: str) -> Tuple[str, str, str]:
    return (
        _parse_genome(composition),
        _parse_type(composition),
        _parse_sense(composition),
    )


def _parse_genome(composition: str) -> str:
    if "DNA" in composition:
        return "DNA"
    elif "RNA" in composition:
        return "RNA"
    else:
        return ""


def _parse_type(composition: str) -> str:
    if "ss" in composition:
        type_ = "single-stranded"
    elif "ds" in composition:
        type_ = "double-stranded"
    else:
        type_ = ""

    if "RT" in composition:
        type_ = type_ + "-RT"

    return type_


def _parse_sense(composition: str) -> str:
    if "+" in composition and "-" in composition:
        return "+/-"
    elif "+" in composition:
        return "+"
    elif "-" in composition:
        return "-"
    else:
        return ""


@functools.lru_cache()
def get_host_aids():
    hosts = get_hosts()

    return [h["aid"] for h in hosts]


@functools.lru_cache()
def get_host_metadata():
    logging.info("Getting metadata...")
    metadata_path = (
        Path(os.path.realpath(__file__)).parent.parent.parent
        / "cache"
        / "host_metadata.json"
    )

    if metadata_path.exists():
        return {
            key: defaultdict(str, value)
            for key, value in json.load(metadata_path.open("r")).items()
        }

    hosts = get_hosts()
    ftp_paths = parse_to_ftp_paths(hosts)

    file_paths = [to_fs_path(path, "coding") for path in ftp_paths]

    records = (
        (_join_sequences(path, host["aid"]), host)
        for path, host in zip(file_paths, hosts)
    )

    metadata = {
        host["aid"]: _host_meta(host["aid"], seq, host["species"])
        for seq, host in records
    }

    json.dump(metadata, metadata_path.open("w"))

    return metadata


def _host_meta(aid: str, record: SeqRecord, species: str):
    sequence = str(record.seq)
    tax = taxonomy.get_ncbi_taxonomy(species)

    meta = {
        "model": "VLMC",
        "name": species,
        "species": species,
        "unique_id": aid,
        "sequence_length": len(sequence),
        "average_gc_content": SeqUtils.GC(sequence) / 100,
        "aid": aid,
    }

    if tax is not None:
        meta = {**meta, **tax}

    return meta


def get_joined_host_coding_sequence() -> Iterable[tuple[SeqRecord, dict[str, str]]]:
    hosts = get_hosts()
    ftp_paths = parse_to_ftp_paths(hosts)

    file_paths = [to_fs_path(path, "coding") for path in ftp_paths]

    return (
        (_join_sequences(path, host["aid"]), host)
        for path, host in zip(file_paths, hosts)
        if path is not None and path.is_file()
    )


def _join_sequences(path: Path, aid: str) -> SeqRecord:
    if path is not None and path.is_file():
        records = SeqIO.parse(gzip.open(path, "rt"), "fasta")
        sequence = "N".join(str(record.seq) for record in records)
        return SeqRecord(Seq(sequence), id=aid)
    else:
        return SeqRecord(Seq(""), id=aid)


@numba.njit
def cosine_distance(left: np.ndarray, right: np.ndarray) -> float:
    denom = np.linalg.norm(left) * np.linalg.norm(right)
    if denom == 0:
        return np.nan

    correlation = np.sum(left * right) / denom

    return 2 * np.arccos(correlation) / np.pi



import functools
import gzip
import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import Dict, Final, Iterable, NewType, Tuple

import numba
import numpy as np

# noinspection PyPackageRequirements
from Bio import SeqIO

# noinspection PyPackageRequirements
from Bio.SeqRecord import SeqRecord

from . import codon
from .codon import CodonTable, VALID_CODONS, _SYNONYMOUS_CODONS
from .common import cosine_distance

CodonPairTable = NewType("CodonPairTable", dict)
CodonPairScoreTable = NewType("CodonPairScoreTable", dict)
AminoAcidTable = NewType("AminoAcidTable", dict)
AminoAcidPairTable = NewType("AminoAcidPairTable", dict)


def generate_codon_pair_score_table(
    codon_counts: CodonTable,
    codon_pair_counts: CodonPairTable,
) -> CodonPairScoreTable:
    """Generate codon pair score table from the codon and codon pair counts.

    Defined for each codon pair as the frequency of the codon pair normalised by
    the frequency of each individual codon and the frequencies of the amino acid pair.

    Parameters
    ----------
    codon_counts: CodonTable
        Codon counts.
    codon_pair_counts: CodonPairTable
        Codon pair counts.

    Returns
    -------
    CodonPairScoreTable
        Codon pair score table.

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    aa_counts = _count_amino_acids(codon_counts)

    aa_pair_counts = _count_amino_acid_pairs(codon_pair_counts)

    cpst = defaultdict(float)

    for codon_pair, counts in codon_pair_counts.items():
        first_codon, second_codon = _split_codon_pair(codon_pair)
        if len(first_codon) == 0 or len(second_codon) == 0:
            continue

        first_aa, second_aa = (
            _CODON_AA_TABLE[first_codon],
            _CODON_AA_TABLE[second_codon],
        )
        aa_pair = first_aa + second_aa

        codon_frequency_normalisation = (
            codon_counts[first_codon]
            * codon_counts[second_codon]
            / (aa_counts[first_aa] * aa_counts[second_aa])
        ) * aa_pair_counts[aa_pair]

        cpst[codon_pair] = np.log(counts / codon_frequency_normalisation)

    return cpst


def _count_amino_acid_pairs(codon_pair_counts: CodonPairTable) -> AminoAcidPairTable:
    aa_pair_counts = defaultdict(float)
    for codon_pair, counts in codon_pair_counts.items():
        first_codon, second_codon = _split_codon_pair(codon_pair)

        if len(first_codon) == 0 or len(second_codon) == 0:
            continue

        first_aa, second_aa = (
            _CODON_AA_TABLE[first_codon],
            _CODON_AA_TABLE[second_codon],
        )

        aa_pair = first_aa + second_aa
        aa_pair_counts[aa_pair] += counts

    return aa_pair_counts


def _count_amino_acids(codon_counts: CodonTable) -> AminoAcidTable:
    aa_counts = defaultdict(float)
    for aa, synonymous_codons in _SYNONYMOUS_CODONS.items():
        aa_counts[aa] = sum(codon_counts[codon] for codon in synonymous_codons)
    return aa_counts


def count_codon_pairs(records: Iterable[SeqRecord]) -> CodonPairTable:
    """Counts the codon pairs in records.  Assumed to be in-frame.

    Parameters
    ----------
    records: Iterable[SeqRecord]
        Record to count codon pairs in.

    Returns
    -------
    CodonPairTable
        Counts of the codon pairs

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    codon_pairs = defaultdict(float)

    for record in records:
        for codon_pair in _codon_pair_generator(record):
            codon_pairs[codon_pair] += 1

    return codon_pairs


_illegal_codon_pairs_discovered = defaultdict(set)


def _codon_pair_generator(record: SeqRecord) -> Iterable[str]:
    sequence = str(record.seq)
    for codon_pair_start in range(0, len(sequence) - 3, 3):
        codon_pair_end = codon_pair_start + 6
        codon_pair = sequence[codon_pair_start:codon_pair_end]

        first_codon, second_codon = _split_codon_pair(codon_pair)
        if first_codon not in VALID_CODONS or second_codon not in VALID_CODONS:
            if codon_pair not in _illegal_codon_pairs_discovered[record.id]:
                logging.warning(f"Illegal codon pair {codon_pair} in gene: {record.id}")
                _illegal_codon_pairs_discovered[record.id].add(codon_pair)
            continue

        yield codon_pair


def _split_codon_pair(codon_pair: str) -> Tuple[str, str]:
    return codon_pair[:3], codon_pair[3:]


def codon_pair_bias_from_counts(
    cpst: CodonPairScoreTable, codon_pair_count: CodonPairTable
) -> float:
    """Calculates the codon pair bias for a given codon pair count.

    Defined as the arithmetic mean of the codon pair scores of the codon pairs.

    Parameters
    ----------
    cpst: CodonPairScoreTable
        The codon pair score table for each codon pair.
    codon_pair_count: CodonPairTable
        Codon pairs to score.

    Returns
    -------
    float
        Codon pair bias.

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    n_codon_pairs = 0
    codon_pair_sum = 0
    for codon_pair, count in codon_pair_count.items():
        codon_pair_sum += cpst[codon_pair] * count
        n_codon_pairs += count

    return codon_pair_sum / (n_codon_pairs - 1)


@numba.njit
def codon_pair_bias_from_counts_(
    cpst: np.ndarray, codon_pair_count: np.ndarray
) -> float:
    """Calculates the codon pair bias for a given codon pair count.

    Defined as the arithmetic mean of the codon pair scores of the codon pairs.

    Parameters
    ----------
    cpst: np.ndarray
        The codon pair score table for each codon pair.
    codon_pair_count: np.ndarray
        Codon pairs to score.

    Returns
    -------
    float
        Codon pair bias.

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    n_codon_pairs = 0
    codon_pair_sum = 0
    for codon_pair_i, count in enumerate(codon_pair_count):
        codon_pair_sum += cpst[codon_pair_i] * count
        n_codon_pairs += count

    return codon_pair_sum / (n_codon_pairs - 1)


def codon_pair_bias(cpst: CodonPairScoreTable, records: Iterable[SeqRecord]) -> float:
    """Calculates the codon pair bias for a gene.

    Defined as the arithmetic mean of the codon pair scores of the codon pairs in records.

    Parameters
    ----------
    cpst: CodonPairScoreTable
        The codon pair score table for each codon pair.
    records: Iterable[SeqRecord]
        Records to score.

    Returns
    -------
    float
        Codon pair bias for the records.

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    n_codon_pairs = 0
    codon_pair_sum = 0
    for record in records:
        for codon_pair in _codon_pair_generator(record):
            codon_pair_sum += cpst[codon_pair]
            n_codon_pairs += 1

    return codon_pair_sum / (n_codon_pairs - 1)


def cache_codon_pair_count(
    aid: str, record_path: Path, cache_dir: Path = Path("../cache/codon-pairs")
) -> CodonPairTable:
    """Caches the codon pair counts, or fetches them from disk if previously calculated.

    Parameters
    ----------
    aid: str
        Unique identifier for the records.
    record_path: Path
        Path to record.
    cache_dir: Path
        Path to where codon counts are cached.

    Returns
    -------
    CodonPairTable
        Count of codon pairs in record_path.

    References
    ----------
    [1] https://doi.org/10.1126/science.1155761
    """
    cache_dir.mkdir(exist_ok=True, parents=True)
    cache_path = (cache_dir / f"{aid}-codon-pairs").with_suffix(".json")

    if cache_path.is_file():
        with cache_path.open("r") as f:
            codon_pair_count = json.load(f)
    else:
        logging.info(f"Counting codon pairs for {aid}")
        if record_path.suffix == ".gz":
            with gzip.open(record_path, "rt") as f:
                records = SeqIO.parse(f, "fasta")
                codon_pair_count = count_codon_pairs(records)
        else:
            with record_path.open("r") as f:
                records = SeqIO.parse(f, "fasta")
                codon_pair_count = count_codon_pairs(records)

        with cache_path.open("w") as f:
            json.dump(codon_pair_count, f)

    return codon_pair_count


@functools.cache
def cache_codon_pair_count_array(aid: str, record_path: Path) -> np.ndarray:
    codon_pair_count = cache_codon_pair_count(aid, record_path)

    return np.array(
        [
            codon_pair_count.get(first_codon + second_codon, 0.0)
            for first_codon in codon.VALID_CODONS
            for second_codon in codon.VALID_CODONS
        ]
    )


@functools.cache
def get_codon_pair_score_table(aid: str, path: Path, cache_path: Path = Path("../cache/codons-and-pairs")) -> CodonPairScoreTable:
    codon_counts = codon.cache_codon_count(aid, path, cache_path)
    codon_pair_counts = cache_codon_pair_count(aid, path, cache_path)
    cpst = generate_codon_pair_score_table(codon_counts, codon_pair_counts)
    return cpst


@functools.cache
def get_codon_pair_score_array(aid: str, path: Path) -> np.ndarray:
    codon_counts = codon.cache_codon_count(aid, path)
    codon_pair_counts = cache_codon_pair_count(aid, path)
    cpst = generate_codon_pair_score_table(codon_counts, codon_pair_counts)

    return np.array(
        [
            cpst.get(first_codon + second_codon, 0.0)
            for first_codon in codon.VALID_CODONS
            for second_codon in codon.VALID_CODONS
        ]
    )


class CPB:
    """Calculate distance as the codon pair bias.

    As defined by Coleman et al. in https://doi.org/10.1126/science.1155761

    Assume sequences gives are in frame.
    """

    def __init__(self, path_dict):
        """
        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    def distance(self, left: str, right: str) -> float:
        cpst = get_codon_pair_score_array(right, self.path_dict[right])

        left_codon_pair_counts = cache_codon_pair_count_array(
            left, self.path_dict[left]
        )

        return codon_pair_bias_from_counts_(cpst, left_codon_pair_counts)

    def transform(self, aid):
        return aid


class CPSCorrelation:
    """Calculate distance as the cosine distance of codon pair scores.

    Codon pair scores are as defined by Coleman et al. in https://doi.org/10.1126/science.1155761

    This assumes sequences gives are in frame.
    """

    def __init__(self, path_dict: Dict[str, Path] = None):
        """

        Parameters
        ----------
        path_dict: Dict[str, Path]
            Dictionary for paths, to load sequences when needed.
        """
        self.path_dict = path_dict

    @functools.singledispatchmethod
    def distance(self, left, right):
        return 0.0

    @distance.register
    def _(self, left: str, right: str):
        if right not in self.path_dict or self.path_dict[right] is None:
            logging.info(f"{right} not in path_dict, or is None")
            return np.inf
        elif left not in self.path_dict or self.path_dict[left] is None:
            logging.info(f"{left} not in path_dict, or is None")
            return np.inf

        left_cpst = get_codon_pair_score_array(left, self.path_dict[left])
        right_cpst = get_codon_pair_score_array(right, self.path_dict[right])

        return self.distance(left_cpst, right_cpst)

    @distance.register
    def _(self, left_p: Path, right_p: Path):
        left_cpst = get_codon_pair_score_array(left_p.stem, left_p)
        right_cpst = get_codon_pair_score_array(right_p.stem, right_p)
        return self.distance(left_cpst, right_cpst)

    @distance.register(dict)
    def _(
        self,
        left_cpst: CodonPairScoreTable,
        right_cpst: CodonPairScoreTable,
    ):

        right_cps_array = np.array(
            [
                right_cpst[first_codon + second_codon]
                for first_codon in codon.VALID_CODONS
                for second_codon in codon.VALID_CODONS
            ]
        )
        left_cps_array = np.array(
            [
                left_cpst[first_codon + second_codon]
                for first_codon in codon.VALID_CODONS
                for second_codon in codon.VALID_CODONS
            ]
        )

        return cosine_distance(left_cps_array, right_cps_array)

    @distance.register
    def _(
        self,
        left_cpst: np.ndarray,
        right_cpst: np.ndarray,
    ):

        return cosine_distance(left_cpst, right_cpst)

    def transform(self, aid) -> np.ndarray:
        """Transforms the aid into the corresponding array.

        Parameters
        ----------
        aid: str
            Aid to transform

        Returns
        -------
        np.ndarray
            Array of codon pair scores, in a deterministic order
        """

        return get_codon_pair_score_array(aid, self.path_dict[aid])


_CODON_AA_TABLE: Final = {
    codon: aa
    for aa, synonymous_codons in _SYNONYMOUS_CODONS.items()
    for codon in synonymous_codons
}



import enum
import functools
import gzip
import http.client
import http.client
import itertools
import json
import logging
import os
import re
import shutil
import subprocess
import urllib
from collections import defaultdict
from pathlib import Path
from typing import Optional, Tuple, Union
from collections.abc import Iterable, Iterator

import Bio
import numpy as np
from Bio import Entrez, SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

import tqdm


def get_or_download_fasta(
    aid: str, download_folder: Path, return_path: bool
) -> Optional[Bio.SeqRecord.SeqRecord | Path]:
    """Gets the fasta file for `aid` either in `download_folder` or by download it.

    Parameters
    ----------
    aid : str
        AID for entry to get fasta for
    download_folder : Path
        Path to directory where fasta files are saved
    return_path: bool
        True to get path, False to get SeqRecord

    Returns
    -------
    Union[Bio.SeqRecord, Path]
        Fasta file contents
    """

    path = get_path(aid, download_folder, [".fa", ".fasta", ".fna"])
    if path is None:
        with get_download_cache() as download_cache:
            if download_cache.has_full_failed(aid):
                return None

            logging.info(f"Found no fasta file for {aid}, attempting download.")
            try:
                fastas = list(get_fastas([aid]))
            except (
                urllib.error.HTTPError,
                http.client.IncompleteRead,
                ValueError,
            ) as e:
                logging.info("Could not download fasta.")
                download_cache.mark_full_as_failed(aid)
                return None

            if fastas is None:
                logging.info("Could not download fasta.")
                return None

            path = download_folder / f"{aid}.fa.gz"
            with gzip.open(path, "wt") as f:
                for fasta in fastas:
                    f.write(fasta.format("fasta"))

    if return_path:
        return path
    else:
        if path.suffix == ".gz":
            with gzip.open(path, "rt") as f:
                return next(SeqIO.parse(f, "fasta"))
        else:
            with path.open("r") as f:
                return next(SeqIO.parse(f, "fasta"))


class DownloadType(enum.Enum):
    Full = "Full"
    Coding = "Coding"


def batch_download(
    ids: list[str],
    download_folder: Path,
    batch_size: int,
    download_type: DownloadType = DownloadType.Full,
):
    if len(ids) == 0:
        return

    logging.info(f"Downloading {ids}, with {download_type}")

    n_batches = np.maximum(len(ids) // batch_size, 1)
    batches = np.array_split(ids, n_batches)

    with tqdm.tqdm(batches) as pbar:
        for batch in pbar:
            pbar.set_description(f"Downloading {batch}")

            try:
                fastas = get_fastas(batch, download_type)
                if fastas is None:
                    return

                suffix = ""
                if download_type == DownloadType.Coding:
                    suffix = "_cds"

                for fasta in fastas:
                    id_ = fasta.id
                    if download_type == DownloadType.Coding:
                        id_, _ = fasta.id.split("_cds_")
                        id_ = id_.replace("lcl|", "")

                    path = download_folder / f"{id_}{suffix}.fa.gz"
                    with gzip.open(path, "at") as f:
                        f.write(fasta.format("fasta"))
            except http.client.IncompleteRead as e:
                logging.warning(e)
                continue


def get_or_download_gb(aid: str, download_folder: Path) -> Optional[SeqRecord]:
    """Gets the genbank file for `aid` either in `download_folder` or by download it.

    Parameters
    ----------
    aid : str
        AID for the entry
    download_folder : Path
        Path to directory where genbank files are saved

    Returns
    -------
    Bio.SeqRecord
        Genbank file.
    """

    gbs = get(aid, download_folder, [".gb"], "genbank")
    if gbs is None:
        logging.info(f"Found no genbank file for {aid}, attempting download.")
        gbs = list(get_gbs([aid]))
        if len(gbs) == 0:
            logging.info("Could not download genbank file.")
            return None

        path = download_folder / f"{aid}.gb.gz"
        with open(path, "w") as f:
            f.write(gbs[0].format("genbank"))

    return gbs[0]


def get(
    aid: str, download_folder: Path, suffixes: list[str], format_name: str = ""
) -> Optional[list[SeqRecord]]:
    """Try to retrieve the file for the aid with a suffix in suffixes.

    Parameters
    ----------
    aid : str
        AID for entry to get file for.
    download_folder : Path
        Path to directory with downloaded files.
    suffixes : list[str]
        list of file endings for downloaded files,
         e.g. [".fa", ".fasta"] for fasta files.
    format_name : str, optional
        Name of format for passing to SeqIO.parse (default "")

    Returns
    -------
    Bio.SeqRecord
        File contents or None if no file found.
    """
    path = get_path(aid, download_folder, suffixes)
    if path is None or not path.is_file():
        return None

    if path.suffix == ".gz":
        try:
            with gzip.open(path, "rt") as f:
                return list(SeqIO.parse(f, format_name))
        except gzip.BadGzipFile as e:
            # TODO oops no compression?
            path_no_comp = path.with_suffix(".tmp")
            shutil.move(path, path_no_comp)
            with gzip.open(path, "wt") as f_out:
                with path_no_comp.open("r") as f_in:
                    f_out.write(f_in.read())
            with gzip.open(path, "rt") as f:
                return list(SeqIO.parse(f, format_name))

    else:
        with path.open("r") as f:
            if ".tax" == path.suffix:
                return json.load(f)
            else:
                return list(SeqIO.parse(f, format_name))


def get_path(aid: str, download_folder: Path, suffixes: list[str]) -> Optional[Path]:
    """Get path to file with `aid` in `download_folder` and suffix in `suffixes`.

    Parameters
    ----------
    aid : str
        Accession id
    download_folder: Path
        Folder/directory where downloads are stored
    suffixes : list[str]
        Possible suffixes for this type of file.

    Returns
    -------
    Path
    """
    download_folder = Path(download_folder)

    # Add compression suffixes
    suffixes = suffixes + [f"{suf}.gz" for suf in suffixes]

    has_version_number = len(aid.split(".")) == 2
    if has_version_number:
        aids = [aid]
    else:
        aids = itertools.chain([aid], (f"{aid}.{i}" for i in reversed(range(50))))

    for aid_ in aids:
        for suffix in suffixes:
            path = download_folder / f"{aid_}{suffix}"

            if path.is_file():
                if path.suffix == ".gz":
                    return path
                else:
                    with gzip.open(f"{path}.gz", "wt") as f_out:
                        with path.open("r") as f_in:
                            f_out.write(f_in.read())
                    path.unlink()
                    return Path(f"{path}.gz")

    return None


class SequencePart(enum.Enum):
    Full = "Full"
    Coding = "Coding"
    NonCoding = "NonCoding"
    DustMasked = "DustMasked"
    TandemRepeatsFinder = "TandemRepeatsFinder"
    Genes = "Genes"


def _get_sequences(p: Path):
    with gzip.open(p, "rt") as f:
        return list(SeqIO.parse(f, "fasta"))


def get_sequence_batch(
    aids: list[str],
    download_folder: Path = Path.home() / "data" / "downloads",
    sequence_part: SequencePart = SequencePart.Full,
    return_path: bool = False,
    batch_size: int = 50,
) -> Iterable[SeqRecord | Path]:
    """Retrieves the sequence with `aid` from `download_folder` or NCBI, in batches.

    Fetches the full sequence, coding, non-coding, or dustmasked sequence according to the `sequence_part` parameter.

    Parameters
    ----------
    aids : list[str]
        Accession ids from NCBi.
    download_folder : Path, optional
        Path to download folder where sequences are saved.  (defaults to "downloads").
    sequence_part : str, optional
        One of "Full", "Coding", "NonCoding", "RepeatMasked", "DustMasked".
    return_path : bool, optional
        False to get SeqRecord, True to get path to file
    batch_size: int, optional
        Number of records to download for each batch.  If the downloads fail frequently, lower the batch size.

    Returns
    -------
    Iterable[SeqRecord | Path]
        Iterable of SeqRecord of the sequence or paths.  Is empty if no sequence is found.

    Raises
    ------
    ValueError
        If sequence_part is not one of "Full", "Coding", "NonCoding", "DustMasked".

    """
    download_folder = Path(download_folder)

    if sequence_part not in SequencePart:
        allowed_values = ", ".join([part.name for part in SequencePart])
        raise ValueError(
            f"sequence_part is not a correct value ({sequence_part}). Should be one of {allowed_values}."
        )

    if sequence_part == SequencePart.Full:
        paths = [get_path(aid, download_folder, [".fa", ".fasta"]) for aid in aids]
        not_found_aids = [aid for aid, p in zip(aids, paths) if p is None]

        batch_download(not_found_aids, download_folder, batch_size, DownloadType.Full)
        all_paths = (get_path(aid, download_folder, [".fa", ".fasta"]) for aid in aids)
        if return_path:
            return all_paths
        else:
            return (SeqIO.read(p, "fasta") for p in all_paths)

    elif sequence_part == SequencePart.Coding:
        paths = [Path(download_folder / f"{aid}_joined_cds.fa.gz") for aid in aids]

        not_found_aids = [aid for aid, p in zip(aids, paths) if not p.is_file()]

        batch_download(not_found_aids, download_folder, batch_size, DownloadType.Coding)

        for aid in not_found_aids:
            try:
                join_cds(aid, download_folder)
            except ValueError as e:
                print(e)
                continue

        all_paths = [
            p
            for aid in aids
            if (p := Path(download_folder / f"{aid}_joined_cds.fa.gz")).is_file()
        ]

        if return_path:
            return all_paths
        else:
            return (_get_sequences(p) for p in all_paths)

    elif sequence_part == SequencePart.Genes:
        paths = [Path(download_folder / f"{aid}_cds.fa.gz") for aid in aids]

        not_found_aids = [aid for aid, p in zip(aids, paths) if not p.is_file()]

        batch_download(not_found_aids, download_folder, batch_size, DownloadType.Coding)

        working_paths = [p for p in paths if p.is_file()]

        if return_path:
            return working_paths
        else:
            return (_get_sequences(p) for p in working_paths)

    elif sequence_part == SequencePart.NonCoding:
        return [non_coding_sequence(aid, download_folder, return_path) for aid in aids]

    elif sequence_part == SequencePart.DustMasked:
        paths = [
            get_path(
                aid,
                download_folder,
                [".fa.dustmasked", ".fasta.dustmasked"],
            )
            for aid in aids
        ]
        for path, aid in zip(paths, aids):
            if path is None:
                record = get_or_download_fasta(aid, download_folder, return_path=False)
                dustmasked = dustmask_record(record)
                path = download_folder / f"{aid}.fa.dustmasked"
                SeqIO.write(dustmasked, path, "fasta")

    elif sequence_part == SequencePart.TandemRepeatsFinder:
        paths = [
            get_path(
                aid,
                download_folder,
                [".fa.2.7.7.80.10.50.500.mask", ".fasta.2.7.7.80.10.50.500.mask"],
            )
            for aid in aids
        ]
        for path, aid in zip(paths, aids):
            if path is None:
                record = get_or_download_fasta(aid, download_folder)
                repeat_masked = tandem_repeats_finder(
                    record, aid, download_folder
                )  # This stores the fasta automatically

                path = get_path(
                    aid,
                    download_folder,
                    [".fa.2.7.7.80.10.50.500.mask", ".fasta.2.7.7.80.10.50.500.mask"],
                )

    if return_path:
        return paths
    else:
        sequences = [SeqIO.parse(path, format="fasta") for path in paths]
        return [SeqRecord(seq=sequence.seq, id=aid, name=aid) for sequence in sequences]


def get_sequence(
    aid: str,
    download_folder: Path = Path.home() / "data" / "downloads",
    sequence_part: SequencePart = SequencePart.Full,
    return_path: bool = False,
) -> Optional[SeqRecord | Path]:
    """Retrieve the sequence with `aid` from `download_folder` or GenBank.

    Fetches the full sequence, coding, non-coding, or dustmasked sequence according to the `sequence_part` parameter.

    Parameters
    ----------
    aid : str
        Accession id from GenBank.
    download_folder : Path, optional
        Path to download folder where sequences are saved.  (defaults to "downloads").
    sequence_part : str, optional
        One of "Full", "Coding", "NonCoding", "RepeatMasked", "DustMasked".
    return_path : bool, optional
        False to get SeqRecord, True to get path to file

    Returns
    -------
    Optional[SeqRecord]
        SeqRecord of the sequence, or None if no sequence is found.

    Raises
    ------
    ValueError
        If sequence_part is not one of "Full", "Coding", "NonCoding", "DustMasked".

    """
    if sequence_part == SequencePart.Full:
        return get_or_download_fasta(aid, download_folder, return_path)
    elif sequence_part == SequencePart.Genes:
        path = Path(download_folder / f"{aid}_cds.fa.gz")

        if not path.is_file():
            # get_or_download_cds(aid, download_folder)
            batch_download([aid], download_folder, 1, DownloadType.Coding)

        if return_path:
            return path
        else:
            return _read_fasta(path)[0]

    elif sequence_part == SequencePart.Coding:
        return coding_sequence(aid, download_folder, return_path)
    elif sequence_part == SequencePart.NonCoding:
        return non_coding_sequence(aid, download_folder, return_path)
    elif sequence_part == SequencePart.DustMasked:
        path = get_path(
            aid,
            download_folder,
            [".fa.dustmasked", ".fasta.dustmasked"],
        )
        if path is None:
            record = get_or_download_fasta(aid, download_folder, return_path=False)
            dustmasked = dustmask_record(record)
            path = download_folder / f"{aid}.fa.dustmasked.gz"
            with gzip.open(path, "wt") as f:
                SeqIO.write(dustmasked, f, "fasta")

    elif sequence_part == SequencePart.TandemRepeatsFinder:
        path = get_path(
            aid,
            download_folder,
            [".fa.2.7.7.80.10.50.500.mask", ".fasta.2.7.7.80.10.50.500.mask"],
        )
        if path is None:
            record = get_or_download_fasta(aid, download_folder, return_path=False)
            _ = tandem_repeats_finder(
                record, aid, download_folder
            )  # This stores the fasta automatically

            path = get_path(
                aid,
                download_folder,
                [".fa.2.7.7.80.10.50.500.mask", ".fasta.2.7.7.80.10.50.500.mask"],
            )
    else:
        allowed_values = ", ".join([part.name for part in SequencePart])
        raise ValueError(
            f"sequence_part is not a correct value ({sequence_part}). Should be one of {allowed_values}."
        )

    if return_path:
        return path
    else:
        sequences = _read_fasta(path)
        return SeqRecord(seq=sequences[0].seq, id=aid, name=aid)


def _read_fasta(path: Path) -> list[SeqRecord]:
    if path.suffix == ".gz":
        with gzip.open(path, "rt") as f:
            sequences = list(SeqIO.parse(f, format="fasta"))
    else:
        with path.open("r") as f:
            sequences = list(SeqIO.parse(f, format="fasta"))
    return sequences


def get_cds_path(aid: str, download_folder: Path) -> Optional[Path]:
    """Get path to CDS.

    Parameters
    ----------
    aid : str
        Accession id in GenBank.
    download_folder : Path
        Path to where cds is saved.

    Returns
    -------
    Optional[Path]
        Path to CDS file.
    """
    download_folder = Path(download_folder)
    file_name = f"{aid}_cds"
    cds_path = get_path(file_name, download_folder, [".fa", ".fasta"])
    return cds_path


def get_or_download_cds(
    aid: str, download_folder: Path, return_path: bool = False
) -> Optional[list[SeqRecord] | Path]:
    """Downloads or fetches coding sequence wth `aid` from GenBank.

    Parameters
    ----------
    aid: str
        Accession id in GenBank
    download_folder: str
        Path to where to save the sequence.
    return_path: bool
        Flag for returning SeqRecord or Path to SeqRecord

    Returns
    -------
    Optional[list[SeqRecord]]
        list of SeqRecord of coding sequences.
    """

    file_name = f"{aid}_cds"

    path = get_path(file_name, download_folder, [".fa", ".fasta"])

    if path is not None and return_path:
        return path

    cds = get(file_name, download_folder, [".fa", ".fasta"], "fasta")

    if cds is None:
        term = f'"{aid}"[Accession]'
        search_handle = Entrez.esearch(
            db="Nucleotide", retmax=1, term=term, idtype="acc"
        )
        search_record = Entrez.read(search_handle)
        ids = search_record.get("Idlist", search_record.get("IdList", []))

        cds_handles = Entrez.efetch(
            db="Nucleotide", id=ids, rettype="fasta_cds_na", retmode="text"
        )
        cds = list(SeqIO.parse(cds_handles, "fasta"))

        path = download_folder / f"{file_name}.fa.gz"
        with gzip.open(path, "wt") as f:
            SeqIO.write(cds, f, "fasta")

    if return_path:
        return path
    else:
        return cds


def get_gbs(ids: list[str]) -> Iterator[SeqRecord]:
    """Get the corresponding genbank files from genbank.

    Uses Entrez to find the correct matches.

    Parameters
    ----------
    ids : list[str]
        list of ids to search for.

    Returns
    -------
    Iterator[SeqRecord]
        The corresponding genbank records.
    """
    try:
        gb_handles = Entrez.efetch(
            db="Nucleotide", id=ids, rettype="gb", retmode="text"
        )

        for gb in SeqIO.parse(gb_handles, "genbank"):
            yield gb
    except (
        http.client.IncompleteRead,
        urllib.error.HTTPError,
        ConnectionResetError,
        RuntimeError,
    ) as e:
        logging.warning(f"Some of ids {ids} could not be downloaded, with {e}.")
        return


def get_fastas(
    ids: list[str], download_type: DownloadType = DownloadType.Full
) -> Iterator[SeqRecord]:
    """Get the corresponding fasta files from genbank.

    Uses Entrez to find the correct matches.

    Parameters
    ----------
    ids : list[str]
        list of ids to search for.
    download_type: DownloadType
        Type of download, either DownloadType.Full or DownloadType.Coding.

    Returns
    -------
    Iterator[SeqRecord]
        The corresponding fasta records.
    """
    term = " OR ".join([f'"{id_}"[Accession]' for id_ in ids])
    search_handle = Entrez.esearch(
        db="Nucleotide", retmax=6000, term=term, idtype="acc"
    )
    search_record = Entrez.read(search_handle)
    ids = search_record.get("Idlist", search_record.get("IdList", []))

    if download_type == DownloadType.Full:
        rettype = "fasta"
    elif download_type == DownloadType.Coding:
        rettype = "fasta_cds_na"
    else:
        raise ValueError(f"Download type {download_type} is invalid")

    try:
        fasta_handles = Entrez.efetch(
            db="Nucleotide", id=ids, rettype=rettype, retmode="text"
        )
    except (
        http.client.IncompleteRead,
        urllib.error.HTTPError,
        ConnectionResetError,
        RuntimeError,
    ) as e:
        logging.warning(f"Some of ids {ids} could not be downloaded, with {e}.")
        return None

    return SeqIO.parse(fasta_handles, "fasta")


def coding_sequence(
    aid: str, download_folder: Path, return_path: bool
) -> Optional[SeqRecord | Path]:
    """Get the coding sequence for the aid.

    Joins individual genes with "NNN" to preserve the frame

    Parameters
    ----------
    aid : str
        Accession id for the sequence.
    download_folder : Path
        Path to downloaded files.
    return_path : bool, optional
        False to get SeqRecord, True to get path to file

    Returns
    -------
    Union[SeqRecord, Path]
        SeqRecord of the coding sequence, or path.
    """
    path_no_comp = Path(download_folder / f"{aid}_joined_cds.fa")
    path = Path(f"{path_no_comp}.gz")
    if path_no_comp.is_file():
        with gzip.open(path, "wt") as f_out:
            with path_no_comp.open("r") as f_in:
                f_out.write(f_in.read())
        path_no_comp.unlink()

    if path.is_file() and return_path:
        # TODO remove once the fuckup is resolved
        try:
            with gzip.open(path, "rt") as f:
                f.readline()
        except gzip.BadGzipFile:
            shutil.move(path, path_no_comp)
            with gzip.open(path, "wt") as f_out:
                with path_no_comp.open("r") as f_in:
                    f_out.write(f_in.read())
        return path
    elif path.is_file():
        with gzip.open(path, "rt") as f:
            return SeqIO.read(f, "fasta")

    with get_download_cache() as download_cache:
        if download_cache.has_coding_failed(aid):
            return None
        try:
            cds: list[SeqRecord] = list(get_or_download_cds(aid, download_folder))
        except (urllib.error.HTTPError, http.client.IncompleteRead, ValueError) as e:
            logging.warning(
                f"Coding sequence of {aid} could not be downloaded, with {e}."
            )
            download_cache.mark_coding_as_failed(aid)
            return None

    record = _join_cds(aid, cds, path)

    if return_path and record is not None:
        return path
    else:
        return record


def join_cds(aid: str, download_folder: Path):
    joined_cds_path = Path(download_folder / f"{aid}_joined_cds.fa.gz")

    if joined_cds_path.is_file():
        return

    raw_cds_path = Path(download_folder / f"{aid}_cds.fa")
    if not raw_cds_path.is_file():
        raise ValueError(f"The cds of {aid} is not downloaded.")

    cds = list(SeqIO.parse(raw_cds_path, "fasta"))
    _join_cds(aid, cds, joined_cds_path)

    return joined_cds_path


def _join_cds(aid: str, cds: list[SeqRecord], path: Path) -> Optional[SeqRecord]:
    gene_records = join_records_by_gene(cds, aid)

    if len(gene_records) == 0:
        return None

    sequence = "NNN".join(str(region.seq) for region in gene_records)
    record = SeqRecord(Seq(sequence), id=aid)

    if path.suffix == ".gz":
        with gzip.open(path, "wt") as f:
            f.write(record.format("fasta"))
    else:
        with path.open("w") as f:
            f.write(record.format("fasta"))

    return record


def join_records_by_gene(records: list[SeqRecord], aid: str) -> list[SeqRecord]:
    """Groups the records by their gene (as defined in the description), and joins their sequences.

    Parameters
    ----------
    records: list[SeqRecord]
        The list of genes to join
    aid: str
        Identifier of the records

    Returns
    -------
    list[SeqRecords]
        Each gene is one entry, and each entry is the concatenation of the genes.

    """
    genes = group_records_by_gene(records)

    return [_join_records(gene_records, aid) for gene_records in genes]


def group_records_by_gene(records: list[SeqRecord]) -> list[list[SeqRecord]]:
    """Groups the records by their gene (as defined in the description).

    Parameters
    ----------
    records: list[SeqRecord]
        Records to group by gene.

    Returns
    -------
    list[list[SeqRecord]]
        list of list with the records for each gene.
    """
    genes = defaultdict(list)
    for record in records:
        genes[_parse_gene(record)].append(record)

    return list(genes.values())


def _join_records(records: list[SeqRecord], aid: str) -> SeqRecord:
    sequence = "".join([str(record.seq) for record in records])
    joined_record = SeqRecord(Seq(sequence))
    joined_record.id = aid + "_" + _parse_gene(records[0])
    return joined_record


def non_coding_sequence(
    aid: str, download_folder: Path, return_path: bool
) -> Optional[SeqRecord | Path]:
    """Get the non-coding sequence for the aid.

    Parameters
    ----------
    aid : str
        Accession id for the sequence.
    download_folder : Path
        Path to downloaded files.
    return_path : bool, optional
        False to get SeqRecord, True to get path to file

    Returns
    -------
    SeqRecord | Path
        SeqRecord of the non-coding sequence, or path.
    """
    cds = get_or_download_cds(aid, download_folder)
    if cds is None:
        return None

    fasta = get_or_download_fasta(aid, download_folder, return_path)

    if fasta is None or cds is None:
        return None

    records = get_non_coding_segments(fasta, cds)

    combined_sequence = "NNN".join([str(rec.seq) for rec in records])
    record = SeqRecord(Seq(combined_sequence), id=aid)

    path = Path(download_folder / f"{aid}_non_coding.fa")
    with path.open("w") as f:
        f.write(record[0].format("fasta"))

    if return_path:
        return path
    else:
        return record


def get_non_coding_segment_bounds_aid(
    aid: str, download_folder: Path
) -> list[tuple[int, int]]:
    """Get the segment bounds (start and end index) from the sequence with `aid`.

    Parameters
    ----------
    aid : str
        Aid of seqrecord in ncbi
    download_folder: Path
        Where sequences are stored on disk

    Returns
    -------
    list[tuple[int, int]]
        list of bounds of non-coding segments

    """
    sequence = get_or_download_fasta(aid, download_folder, return_path=False)
    cds = get_or_download_cds(aid, download_folder)
    return get_non_coding_segment_bounds(sequence, cds)


def get_non_coding_segment_bounds(
    sequence: SeqRecord, cds: list[SeqRecord]
) -> list[tuple[int, int]]:
    """Get the segment bounds (start and end index) in sequence that is not referenced as coding in the cds.

    Parameters
    ----------
    sequence : SeqRecord
        Full sequence
    cds : list[SeqRecord]
        list of coding sequence records

    Returns
    -------
    list[tuple[int, int]]
        list of non-coding segments

    """
    sequence = str(sequence.seq)

    starts = []
    ends = []
    for region in cds:
        region_starts, region_ends = _get_coding_region_bounds(region)

        starts.extend(region_starts)
        ends.extend(region_ends)

    starts, ends = _merge_overlapping_regions(starts, ends)

    def bounds():
        last_end = 0
        for start, end in zip(starts, ends):
            if start - last_end > 0:
                yield last_end, start
            last_end = end + 1
        yield last_end, len(sequence)

    non_coding_sequences = [(start, end) for start, end in bounds()]

    return non_coding_sequences


def get_non_coding_segments(
    sequence: SeqRecord, cds: list[SeqRecord]
) -> list[SeqRecord]:
    """Get the segments in sequence that is not referenced as coding in the cds.

    Parameters
    ----------
    sequence : SeqRecord
        Full sequence
    cds : list[SeqRecord]
        list of coding sequence records

    Returns
    -------
    list[SeqRecord]
        list of non-coding segments

    """
    bounds = get_non_coding_segment_bounds(sequence, cds)

    non_coding_sequences = [
        SeqRecord(
            Seq(sequence[start:end]),
            id=f"non_coding_{start}-{end}",
            description=f"[locus_tag=non_coding_{start}_{end}]",
        )
        for start, end in bounds
    ]

    return non_coding_sequences


def _merge_overlapping_regions(
    starts: list[int], ends: list[int]
) -> Tuple[list[int], list[int]]:
    merged_starts = []
    merged_ends = []
    for start, end in zip(sorted(starts), sorted(ends)):
        if len(merged_ends) == 0 or start > merged_ends[-1]:
            merged_starts.append(start)
            merged_ends.append(end)
        else:
            merged_ends[-1] = end

    return merged_starts, merged_ends


def _get_coding_region_bounds(region: SeqRecord) -> tuple[list[int], list[int]]:
    starts = []
    ends = []

    # I'm sure there's an easier way to find this information somehow.
    pattern = (
        r"location=[(join)|(complement)\(]*"
        r"(?P<joins>([<>]?[0-9]+[<>]?(\.\.[<>]?[0-9]+[<>]?)?,?)+)\)?"
    )
    m = re.search(pattern, region.description)
    if m is not None:
        groupdict = m.groupdict()
        bounds = groupdict["joins"].split(",")
        for bound in bounds:
            if ".." in bound:
                start, end = bound.split("..")
            else:
                start, end = bound, bound

            starts.append(int(start.replace("<", "").replace(">", "")))
            ends.append(int(end.replace("<", "").replace(">", "")))

    return starts, ends


def _parse_gene(region: SeqRecord) -> str:
    pattern = r"((gene)|(locus_tag))=(?P<gene>[a-zA-Z0-9_]+)"
    m = re.search(pattern, region.description)
    if m is not None:
        groupdict = m.groupdict()
        return groupdict["gene"]
    else:
        return ""


def tandem_repeats_finder(
    record: SeqRecord, aid: str, download_folder: Path
) -> Optional[SeqRecord]:
    """Runs tandem repeats finder on the SeqRecord

    Parameters
    ----------
    record: SeqRecord
        Record to run tandem repeats finder on
    aid: str
        Aid of the record
    download_folder: Path
        Path to where the result and record is stored

    Returns
    -------
    SeqRecord
        Repeat masked sequence
    """
    trf_parameters = ("2", "7", "7", "80", "10", "50", "500", "-f", "-d", "-h", "-m")

    masked_extra_suffix = "2.7.7.80.10.50.500.mask"
    suffixes = ["fasta", "fa"]

    if not any(
        (Path(download_folder) / f"{aid}.{suffix}").is_file() for suffix in suffixes
    ):
        # None of the suffixes worked, so need to write file first.
        p = Path(download_folder) / f"{aid}.fasta"
        SeqIO.write(record, p, "fasta")

    for suffix in suffixes:
        p = Path(download_folder) / f"{aid}.{suffix}"
        if p.is_file():
            args = ("trf", str(p), *trf_parameters)
            res = subprocess.run(args, capture_output=True, cwd=download_folder)
            if res.returncode != 0:
                logging.warning("Tandem Repeats Masker did not work")
                logging.warning(f"Error message: {res.stderr}")

            masked_p = f"{p}.{masked_extra_suffix}"
            rec = SeqIO.read(masked_p, "fasta")
            return rec

    return None


def dustmask_record(record: SeqRecord) -> SeqRecord:
    """Runs dustmasker as on the SeqRecord.

    The dustmasker program has to be installed on the system, and is run as a subprocess.  Can be found on ubuntu
    in the ncbi-tools package.

    Parameters
    ----------
    record: SeqRecord
        Records to dustmask

    Returns
    -------
    SeqRecord
        The `record` with sequence updated as the dustmasked sequence.

    """

    logging.info(f"Running dustmasker on {record.id}")
    args = "/usr/bin/dustmasker -in -".split(" ")
    completed = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        input=record.format("fasta").encode(),
    )
    output = completed.stdout.decode()

    intervals = [
        _parse_interval(line)
        for line in output.split("\n")
        if len(line) > 0 and line[0] != ">"
    ]

    if len(intervals) == 0:
        return record

    sequence = ""
    prev_end = 0
    for start, end in intervals:
        sequence += str(record.seq[prev_end:start]) + "N" * (end - start)
        prev_end = end

    sequence += str(record.seq[prev_end : len(record.seq)])

    new_record = SeqRecord(record)
    new_record.seq = Seq(sequence)
    new_record.id = record.id

    return new_record


def _parse_interval(line: str) -> Tuple[int, int]:
    [start, end] = line.split(" - ")
    return int(start), int(end)


class DownloadCache:
    DOWNLOAD_STATUS_CACHE_PATH: Path = (
        Path.home() / "cache" / "seqhelp" / "download-status.json"
    )
    DOWNLOAD_STATUS_CACHE_PATH.parent.mkdir(parents=True, exist_ok=True)

    _cache = dict()

    def __init__(self):
        if self.DOWNLOAD_STATUS_CACHE_PATH.is_file():
            try:
                with self.DOWNLOAD_STATUS_CACHE_PATH.open("r") as f:
                    self._cache = json.load(f)
            except json.decoder.JSONDecodeError as e:
                logging.info(f"Failed to load download cache status with {e}")
                self._cache = dict()
        else:
            self._cache = dict()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        with self.DOWNLOAD_STATUS_CACHE_PATH.open("w") as f:
            json.dump(self._cache, f)

    def __getitem__(self, item):
        if item in self._cache:
            return self._cache[item]

    def __setitem__(self, key, value):
        self._cache[key] = value

    def mark_coding_as_failed(self, key):
        if key not in self._cache:
            self._cache[key] = dict()

        self._cache[key]["coding"] = False

    def mark_full_as_failed(self, key):
        if key not in self._cache:
            self._cache[key] = dict()

        self._cache[key]["full"] = False

    def has_coding_failed(self, key):
        if key in self._cache and "coding" in self._cache[key]:
            return True
        else:
            return False

    def has_full_failed(self, key):
        if key in self._cache and "full" in self._cache[key]:
            return True
        else:
            return False


@functools.cache
def get_download_cache():
    return DownloadCache()


def _adjacent_ns(left: str, right: str) -> bool:
    return left == "N" and right == "N"


def compress_ns(record: SeqRecord) -> SeqRecord:
    """
    Compresses adjacent N characters in the SeqRecord.
    Parameters
    ----------
    record: SeqRecord from biopython to compress

    Returns
    -------
    SeqRecord with compressed Ns
    """
    sequence = compress_ns_str(str(record.seq))

    new_record = SeqRecord(record)
    new_record.seq = Seq(sequence)
    new_record.id = record.id
    return new_record


def compress_ns_str(sequence: str) -> str:
    """
    Compresses adjacent N characters in the str.
    Parameters
    ----------
    sequence: str to compress

    Returns
    -------
    SeqRecord with compressed Ns
    """
    sequence = "".join(
        [s for s, sn in zip(sequence, sequence[1:]) if not _adjacent_ns(s, sn)]
        + [sequence[-1]]
    )

    return sequence



import functools
import logging
from collections import defaultdict
from pathlib import Path
import urllib.request

import pandas as pd
import os
from typing import List, Dict, Tuple, Set


@functools.lru_cache()
def _get_virus_host_db(path: str = None):
    if path is None:
        _virus_host_db_path = (
            Path(os.path.realpath(__file__)).parent / "virushostdb.tsv"
        )
        path = _virus_host_db_path

    if not Path(path).is_file():
        url = "https://www.genome.jp/ftp/db/virushostdb/virushostdb.tsv"

        logging.info(f"virushostdb not found, trying to fetch from {url}")

        urllib.request.urlretrieve(url, path)

    return _parse_virus_host_db(path)


@functools.lru_cache()
def _parse_virus_host_db(path: str):
    hosts_df = pd.read_csv(path, sep="\t")
    return hosts_df


def get_virus_hosts(
    tax_ids: List[Tuple[int, str, str]],
    path: str = None,
) -> Dict[str, Set[str]]:
    """Finds the virus hosts in a tab separated file by their tax id.

    The tsv file is from  ftp://ftp.genome.jp/pub/db/virushostdb/virushostdb.tsv

    Parameters
    ----------
    tax_ids : List[Tuple[int, str, str]]
        Taxa id, species and refseq id for each virus.
    path : str, optional
        Path to tsv file with viral host information
            Will be downloaded if not found.

    Returns
    -------
    dict
        Dictionary with species as key, and dictionary with metadata as value
        (with classes as key, e.g. "superkingdom", "gc_class")
    """

    hosts_df = _get_virus_host_db(path)

    return {
        species: set(
            host
            for host in _get_hosts(hosts_df, tax_id, species, refseq_aid)
            if str(host) != "nan"
        )
        for tax_id, species, refseq_aid in tax_ids
    }


def _get_hosts(hosts_df, tax_id: int, species: str, refseq_aid: str):
    if tax_id is None:
        tax_id_matches = []
    else:
        tax_id_matches = list(
            hosts_df[hosts_df["virus tax id"] == tax_id]["host name"].values
        )

    if species is None or species == "":
        species_matches = []
    else:
        species_matches = list(
            hosts_df[hosts_df["virus name"].str.contains(species, regex=False)][
                "host name"
            ].values
        ) + list(
            hosts_df[hosts_df["virus lineage"].str.contains(species, regex=False)][
                "host name"
            ].values
        )

    if refseq_aid is None or refseq_aid == "":
        refseq_matches = []
    else:
        refseq_aid = refseq_aid.split(".")[0]  # Don't care about version

        refseq_matches = list(
            hosts_df[hosts_df["refseq id"].str.contains(refseq_aid, regex=False)][
                "host name"
            ].values
        )

    return tax_id_matches + species_matches + refseq_matches


def get_host_names(tax_id: int, species: str, refseq_aid: str, path: str = None) -> List[str]:
    hosts_df = _get_virus_host_db(path)
    hosts = set(_get_hosts(hosts_df, tax_id, species, refseq_aid))
    return [name for name in hosts if str(name) != "nan"]
